/* soundmosaic, Copyright (c) 2001-2007 Steven Hazel <sah@thalassocracy.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */
#ifndef _WIN32
  #include <sys/types.h>
  #include <sys/socket.h>
  #include <errno.h>
  #include <unistd.h>
#else
  #if( !defined( WIN32_LEAN_AND_MEAN ) )
    // Needed to avoid redefinitions by Windows interfaces.
    #define WIN32_LEAN_AND_MEAN
  #endif
  #include <winsock2.h>
  #include <Ws2tcpip.h>
  #include <windows.h>
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#include <string.h>
#include <assert.h>

#include "io_buffer.h"


void io_buffer_reset(io_buffer *self)
{
  self->read_pos = 0;
  self->write_pos = 0;
  self->failed = FALSE;
  memset(self->data, 0xf0, self->size);
}


void io_buffer_init(io_buffer *self, size_t size)
{
  self->data = (unsigned char *)malloc(size);
  if (self->data == NULL) {
    fprintf(stderr,
            "failed to allocate %d bytes for io buffer\n", size);
    exit(0);
  }

  self->size = size;

  io_buffer_reset(self);
}


void io_buffer_destroy(io_buffer *self)
{
  free(self->data);
  self->data = NULL;
}


void io_buffer_grow(io_buffer *self, size_t size)
{
  // growing only, "size" is the total new size (not relative)
  if (size <= self->size) {
    return;
  }

  self->data = realloc(self->data, size);
  if (self->data == NULL) {
    fprintf(stderr,
            "failed to allocate %d bytes for io buffer\n", size);
    exit(0);
  }

  memset(&self->data[self->size + 1], 0xf0, size - self->size);

  self->size = size;
}


void io_buffer_assert_sanity(const io_buffer *self)
{
  assert(self->data != NULL);
  assert(self->size > 0);
  assert(self->read_pos <= self->size);
  assert(self->write_pos <= self->size);
  assert(self->read_pos <= self->write_pos);
}


int io_buffer_is_empty(const io_buffer *self)
{
  if (self->write_pos > self->read_pos) {
    return FALSE;
  } else {
    return TRUE;
  }
}


int io_buffer_is_full(const io_buffer *self)
{
  if ((self->write_pos - self->read_pos) < self->size) {
    return FALSE;
  } else {
    return TRUE;
  }
}


size_t io_buffer_bytes_free(const io_buffer *self)
{
  if (self == NULL) {
    return 0;
  } else {
    return (self->size - self->write_pos);
  }
}


size_t io_buffer_bytes_used(const io_buffer *self)
{
  return self->write_pos - self->read_pos;
}


int io_buffer_failed(const io_buffer *self)
{
  return self->failed;
}


int io_buffer_peek_int(io_buffer *self, size_t len, size_t offset)
{
  unsigned char *data;
  int retval = 0;
  size_t i;

  assert(len <= sizeof(int));

  if ((self->failed) ||
      ((self->read_pos + offset + len) > self->write_pos)) {
    self->failed = TRUE;
    return 0;
  }

  data = &(self->data[self->read_pos + offset]);

  for (i = 0; i < len; i++) {
    retval <<= 8;
    retval += data[i];
  }

  io_buffer_assert_sanity(self);
  return retval;
}


void io_buffer_discard_bytes(io_buffer *self, size_t num)
{
  if (num + self->read_pos > self->write_pos) {
    self->failed = TRUE;
    return;
  }

  self->read_pos += num;

  io_buffer_assert_sanity(self);
  return;
}


void io_buffer_compact(io_buffer *self)
{
  size_t unread_bytes, read_bytes;

  unread_bytes = (self->write_pos - self->read_pos);
  read_bytes = self->read_pos;

  if (read_bytes == 0) {
    return;
  }

  /* move the unread data to the beginning of the buffer */
  memmove(self->data, (self->data + self->read_pos), unread_bytes);

  /* fill in the rest of the gap with nonsense */
  memset((self->data + self->write_pos), 0xf0,
         (self->size - self->write_pos));

  self->write_pos = self->write_pos - read_bytes;
  self->read_pos = 0;

  io_buffer_assert_sanity(self);
  return;
}


ssize_t io_buffer_recv_all(io_buffer *self, SOCKET fd)
{
  ssize_t bytes_read = 0;
  size_t avail_buffer = 0;
  unsigned char *start_addr = NULL;

  io_buffer_compact(self);

  assert(self->size >= self->write_pos);

  avail_buffer = self->size - self->write_pos;
  start_addr = &self->data[self->write_pos];

  if (avail_buffer == 0) {
    return -1;
  }

#if defined(_WIN32) && !defined(_WIN64)
  // I can't imagine this will stick around
  bytes_read = recv(fd, (char *)start_addr, (int)avail_buffer, 0);
#else
  bytes_read = recv(fd, (char *)start_addr, avail_buffer, 0);
#endif
  if (bytes_read <= 0) {
    return bytes_read;
  }

  self->write_pos += bytes_read;

  assert(self->write_pos <= self->size);

  io_buffer_assert_sanity(self);
  return bytes_read;
}


ssize_t io_buffer_send_all(io_buffer *self, SOCKET fd)
{
  ssize_t bytes_sent = 0;
  size_t bytes_to_send = 0;
  unsigned char *start_addr = NULL;

  start_addr = &(self->data[self->read_pos]);

  assert(self->write_pos >= self->read_pos);

  bytes_to_send = self->write_pos - self->read_pos;

  if (bytes_to_send == 0) {
    return 0;
  }

#if defined(_WIN32) && !defined(_WIN64)
  // I can't imagine this will stick around
  bytes_sent = send(fd, (char *)start_addr, (int)bytes_to_send, 0);
#else
  bytes_sent = send(fd, (char *)start_addr, bytes_to_send, 0);
#endif
  if (bytes_sent <= 0) {
    return bytes_sent;
  }

  io_buffer_discard_bytes(self, bytes_sent);
  io_buffer_compact(self);

  io_buffer_assert_sanity(self);
  return bytes_sent;
}


size_t io_buffer_fread_all(io_buffer *self, FILE *stream)
{
  size_t bytes_read = 0;
  size_t avail_buffer = 0;
  unsigned char *start_addr = NULL;

  io_buffer_compact(self);

  assert(self->size >= self->write_pos);

  avail_buffer = self->size - self->write_pos;
  start_addr = &self->data[self->write_pos];

  if (avail_buffer == 0) {
    return 0;
  }

  bytes_read = fread(start_addr, 1, avail_buffer, stream);
  if (bytes_read == 0) {
    return bytes_read;
  }

  self->write_pos += bytes_read;

  assert(self->write_pos <= self->size);

  io_buffer_assert_sanity(self);
  return bytes_read;
}


size_t io_buffer_fread(io_buffer *self, FILE *stream, size_t len)
{
  size_t bytes_read = 0;
  size_t avail_buffer = 0;
  unsigned char *start_addr = NULL;

  io_buffer_compact(self);

  assert(self->size >= self->write_pos);

  avail_buffer = self->size - self->write_pos;
  start_addr = &self->data[self->write_pos];

  if (avail_buffer == 0) {
    self->failed = TRUE;
    return 0;
  }

  if (len > avail_buffer) {
    len = avail_buffer;
  }

  bytes_read = fread(start_addr, 1, len, stream);
  if (bytes_read == 0) {
    return bytes_read;
  }

  self->write_pos += bytes_read;

  assert(self->write_pos <= self->size);

  io_buffer_assert_sanity(self);
  return bytes_read;
}


size_t io_buffer_fwrite(io_buffer *self, FILE *stream, size_t len)
{
  size_t bytes_sent = 0;
  size_t bytes_avail = 0;
  unsigned char *start_addr = NULL;

  start_addr = &(self->data[self->read_pos]);
  bytes_avail = self->write_pos - self->read_pos;

  if (bytes_avail == 0) {
    self->failed = TRUE;
    return 0;
  }

  if (len > bytes_avail) {
    len = bytes_avail;
  }

  bytes_sent = fwrite(start_addr, 1, len, stream);
  if (bytes_sent == 0) {
    return bytes_sent;
  }

  io_buffer_discard_bytes(self, bytes_sent);
  io_buffer_compact(self);

  io_buffer_assert_sanity(self);
  return bytes_sent;
}


void io_buffer_read(io_buffer *self, char *dest, size_t len)
{

  if ((self->failed) ||
      (self->read_pos + len > self->write_pos)) {
    self->failed = TRUE;
    return;
  }

  memcpy(dest, &self->data[self->read_pos], len);
  self->read_pos = self->read_pos + len;

  io_buffer_assert_sanity(self);
  return;
}


void io_buffer_write(io_buffer *self, const char *src, size_t len)
{

  if ((self->failed) ||
      (self->write_pos + len >= self->size)) {
    self->failed = TRUE;
    return;
  }

  memcpy(&self->data[self->write_pos], src, len);
  self->write_pos = self->write_pos + len;

  io_buffer_assert_sanity(self);
  return;
}
