/* soundmosaic, Copyright (c) 2001-2007 Steven Hazel <sah@thalassocracy.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */
#include "config.h"

#include <stdio.h>
#include <signal.h>
#include <limits.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <stdlib.h>
#include <pthread.h>
#include <audiofile.h>

#ifdef HAVE_GETOPT
#include <getopt.h>
#else
#include "getopt/getopt.h"
#endif

#include "main.h"
#include "soundmosaic.h"
#include "sm_utils.h"
#include "resampler.h"
#include "server.h"
#include "client.h"
#include "graphics.h"

#define INTERNAL_BYTEORDER AF_BYTEORDER_LITTLEENDIAN
#define INTERNAL_FORMAT AF_SAMPFMT_TWOSCOMP
#define INTERNAL_WIDTH (sizeof(int16_t) * 8)

typedef struct {
  AFfilehandle file;

  int file_format;
  int format;
  int width;
  int channels;
  int frame_count;
  int bytes;
  int rate;
  int byte_order;

} audio_file_info;

typedef struct {
  int16_t **data;
  int32_t len;
  int channels;
} source_buffer;

typedef struct {
  int16_t *data;
  int frame_count;
  int channels;
} resume_buffer;

int dying = 0;

node *srcs_root = NULL;
node *srcs_end = NULL;

node *attempts_root = NULL;
node *attempts_end = NULL;

void termination_handler (int signum)
{
  dying = 1;
}


void usage (void)
{
  fprintf(stderr,

"soundmosaic " VERSION "\n"
"Copyright (c) 2001-2007 Steven Hazel <sah@thalassocracy.org>\n"
"\n"
"usage:\n"
"\n"
"  soundmosaic [options] <target file> <output file>"
" [at least one source file]\n"
"\n"
"or, to collate existing output files:\n"
"\n"
"  soundmosaic [options] --collate\n"
"              <target file> <output file> [at least two attempt files]\n"
"\n"
"or, to run as a master server, collating output from slaves:\n"
"\n"
"  soundmosaic [options] --master <port>\n"
"              <target file> <output file> [optional local attempt files]\n"
"\n"
"or, to run as a slave:\n"
"\n"
"  soundmosaic [options] --slave <url> [at least one source file]\n"
"\n"
"options:\n"
"\n"
"  -t, --tilesize=SECONDS   Seconds per tile (usually\n"
"                           fractional).  Defaults to 0.1.\n"
"\n"
"  -p, --partition          Merely partition the source file(s)\n"
"                           into tiles, rather than using all of\n"
"                           the continuous (overlapping) tiles in\n"
"                           the source file(s).  This is much\n"
"                           faster, but produces lower quality\n"
"                           results.\n"
"\n"
"  -r, --resume             Attempt to resume processing on an\n"
"                           existing, aborted output file.  Not\n"
"                           valid when collating, or when acting\n"
"                           as a master or slave.\n"
"\n"

          );
}


void print_soundmosaic_error(int error)
{
  switch (error) {
  case SOUNDMOSAIC_SUCCESS:
    fprintf(stderr, "soundmosaic succeeded\n");
    break;
  case SOUNDMOSAIC_INVALID_TILE_SIZE:
    fprintf(stderr, "soundmosaic failed: tile size is invalid\n");
    break;
  default:
    fprintf(stderr, "soundmosaic failed with unknown error code %d\n", error);
  }
}


int copy_stream( FILE *sourcef, FILE *destf )
{
  int status;
  char buf[128];

  do {
    status = fread(buf, 1, 128, sourcef);
    if (status > 0) {
      fwrite(buf, 1, status, destf);
    }
  } while (status == 128);

  if (feof(sourcef)) {
    return TRUE;
  } else {
    return FALSE;
  }
}


void free_source_buffer(source_buffer *this)
{
  int i;

  for (i = 0; i < this->channels; i++) {
    free(this->data[i]);
  }

  free(this->data);
}


int buffer_source_file(source_buffer *this,
                       FILE *src_stream,
                       audio_file_info *tgt)
{
  audio_file_info src;
  int32_t src_len = 0;
  int32_t current_frame;
  int16_t *ibuf = NULL;
  int16_t *src_buf = NULL;
  int16_t *tempbuf = NULL;
  int i;
  int status;

  /* open the source file */
  src.file = afOpenFD(fileno(src_stream), "r", NULL);
  if (!src.file) {
    fprintf(stderr, "couldn't open source file as an audio file\n");
    return FALSE;
  }

  /* get target file parameters */
  src.frame_count = afGetFrameCount(src.file, AF_DEFAULT_TRACK);
  src.channels = afGetChannels(src.file, AF_DEFAULT_TRACK);
  src.rate = afGetRate(src.file, AF_DEFAULT_TRACK);
  src.bytes = afGetTrackBytes(src.file, AF_DEFAULT_TRACK);

  /* make sure data is converted into the preferred format */
  afSetVirtualSampleFormat(src.file, AF_DEFAULT_TRACK, tgt->format,
                           tgt->width);
  afSetVirtualByteOrder(src.file, AF_DEFAULT_TRACK,
                        INTERNAL_BYTEORDER);
  //afSetVirtualRate(src.file, AF_DEFAULT_TRACK, tgt->rate);

  printf("source file is %f seconds long (%d frames)\n",
         ((float)src.frame_count / (float)src.rate), src.frame_count);

  //fprintf(stderr, "frames: %d channels: %d rate: %d seconds: %d\n",
  //        src.frame_count, src.channels, src.rate,
  //        (src.frame_count / src.rate));

  current_frame = 0;

  printf("buffering source file\n");

  src_buf = (int16_t *)malloc(sizeof(int16_t) * src.channels *
                              src.frame_count);
  if (src_buf == NULL) {
    afCloseFile(src.file);
    fprintf(stderr, "failed allocating source data buffer\n");
    return FALSE;
  }

  tempbuf = (int16_t *)malloc(sizeof(int16_t) * src.channels *
                              src.frame_count);
  if (tempbuf == NULL) {
    free(src_buf);
    afCloseFile(src.file);
    fprintf(stderr, "failed allocating temporary buffer\n");
    return FALSE;
  }

  this->data = (int16_t **)malloc(sizeof(int16_t *) * src.channels);
  if (this->data == NULL) {
    free(src_buf);
    free(tempbuf);
    afCloseFile(src.file);
    fprintf(stderr, "failed allocating source data buffer\n");
    return FALSE;
  }

  for (i = 0; i < src.channels; i++) {
    this->data[i] = NULL;
  }

  status = afReadFrames(src.file, AF_DEFAULT_TRACK,
                        src_buf, src.frame_count);
  if (status != src.frame_count) {
    free(src_buf);
    free(tempbuf);
    free_source_buffer(this);
    afCloseFile(src.file);
    fprintf(stderr, "failed reading %d frames from source file, status %d\n",
            src.frame_count, status);
    return FALSE;
  }

  deinterlace(src_buf, tempbuf, src.frame_count * src.channels, src.channels);

  for (i = 0; i < src.channels; i++) {
      ibuf = &tempbuf[i * src.frame_count];

      // resample the channel buffer into a new channel buffer
      resample(src.rate, tgt->rate, ibuf, &(this->data[i]),
               src.frame_count, &src_len);

      if (dying) {
        afCloseFile(src.file);
        free(src_buf);
        free(tempbuf);
        free_source_buffer(this);
        return FALSE;
      }

  }

  if (afCloseFile(src.file)) {
    free(src_buf);
    free(tempbuf);
    free_source_buffer(this);
    fprintf(stderr, "failed closing source file\n");
    return FALSE;
  }

  assert(src_len > 0);
  assert(src.channels > 0);

  this->len = src_len;
  this->channels = src.channels;

  printf("done buffering source file\n");

  free(src_buf);
  free(tempbuf);

  return TRUE;
}


int buffer_resume_stream(FILE *res_stream, resume_buffer *res_buffer)
{
  int status;
  AFfilehandle file;

  /*  try to open the resume file */
  file = afOpenFD(fileno(res_stream), "r", NULL);
  if (file) {

    afSetVirtualByteOrder(file, AF_DEFAULT_TRACK,
                          INTERNAL_BYTEORDER);

    afSetVirtualSampleFormat(file, AF_DEFAULT_TRACK, INTERNAL_FORMAT,
                             INTERNAL_WIDTH);

    res_buffer->frame_count = afGetFrameCount(file, AF_DEFAULT_TRACK);
    res_buffer->channels = afGetChannels(file, AF_DEFAULT_TRACK);

    res_buffer->data = (int16_t *)malloc(sizeof(int16_t) * res_buffer->channels
                                         * res_buffer->frame_count);
    if (res_buffer->data == NULL) {
      return FALSE;
    }

    status = afReadFrames(file, AF_DEFAULT_TRACK,
                          res_buffer->data, res_buffer->frame_count);
    if (status != res_buffer->frame_count) {
      afCloseFile(file);
      free(res_buffer->data);
      return FALSE;
    }
  }

  if (afCloseFile(file)) {
    free(res_buffer->data);
    return FALSE;
  }

  return TRUE;
}


void free_resume_buffer(resume_buffer *res_buffer)
{
  assert(res_buffer != NULL);
  res_buffer->frame_count = 0;
  res_buffer->channels = 0;
  free(res_buffer->data);
  res_buffer->data = 0;

  return;
}


int soundmosaic_streams(FILE *tgt_stream,
                        FILE *src_stream,
                        FILE *out_stream,
                        float tile_seconds,
                        int partition,
                        int graphics,
                        resume_buffer *res_buffer)
{
  int tile_size;
  soundmosaic_state *sm = NULL;
  audio_file_info tgt;
  audio_file_info out;
  AFfilesetup outconfig = NULL;
  int16_t *ibuf = NULL;
  int16_t *obuf = NULL;
  int16_t *iobuf = NULL;
  int16_t *prebuf = NULL;
  int16_t *postbuf = NULL;

  source_buffer src_buffer;

  int frames_read;
  int total_frames_read;
  char format_name[32];
  int i;
  int status;

  int tick_frames;
  int last_mark;
  int start_time = 0;
  int end_time = 0;

  int res_ok = FALSE;
  int retval = FALSE;

  void *draw = NULL;

  tgt.file = NULL;
  out.file = NULL;

  /* init the source_buffer */
  src_buffer.data = NULL;
  src_buffer.len = 0;
  src_buffer.channels = 0;

  /* open the target file */
  tgt.file = afOpenFD(fileno(tgt_stream), "r", NULL);
  if (!tgt.file) {
    fprintf(stderr, "couldn't open target file as an audio file\n");
    retval = FALSE;
    goto end;
  }

  /* get target file parameters */
  tgt.file_format = afGetFileFormat(tgt.file, NULL);
  tgt.frame_count = afGetFrameCount(tgt.file, AF_DEFAULT_TRACK);
  tgt.channels = afGetChannels(tgt.file, AF_DEFAULT_TRACK);
  tgt.rate = afGetRate(tgt.file, AF_DEFAULT_TRACK);
  tgt.bytes = afGetTrackBytes(tgt.file, AF_DEFAULT_TRACK);
  tgt.byte_order = afGetByteOrder(tgt.file, AF_DEFAULT_TRACK);
  afGetSampleFormat(tgt.file, AF_DEFAULT_TRACK, &tgt.format, &tgt.width);

  tile_size = tgt.rate * tile_seconds;
  assert(tile_size >= 0);
  if (tile_size == 0) {
    fprintf(stderr, "tile size too small: less than one sample per frame\n");
    retval = FALSE;
    goto end;
  } else if (tile_size > tgt.frame_count) {
    fprintf(stderr,
            "tile size too large: target file is only %f seconds long\n",
            ((float)tgt.frame_count / (float)tgt.rate));
    retval = FALSE;
    goto end;
  }

  switch (tgt.format) {
  case AF_SAMPFMT_DOUBLE:
    strcpy(format_name, "DOUBLE");
    break;
  case AF_SAMPFMT_UNSIGNED:
    strcpy(format_name, "UNSIGNED");
    break;
  case AF_SAMPFMT_TWOSCOMP:
    strcpy(format_name, "TWOSCOMP");
    break;
  case AF_SAMPFMT_FLOAT:
    strcpy(format_name, "FLOAT");
    break;
  default:
    strcpy(format_name, "UNKNOWN");
  }

  //fprintf(stderr, "target file:\n");

  //fprintf(stderr,
  //      "frames: %d, channels: %d, rate: %d, format: %s, width: %d\n",
  //      tgt.frame_count, tgt.channels, tgt.rate, format_name, tgt.width);

  //fprintf(stderr, "tile size: %d frames, %f seconds\n", tile_size,
  //      ((float)tile_size / (float)tgt.rate));


  if (res_buffer != NULL) {
    if ((res_buffer->frame_count < tgt.frame_count) &&
        (res_buffer->channels == tgt.channels)) {
      res_ok = TRUE;
      printf("continuing aborted processing\n");
    } else if ((res_buffer->frame_count == tgt.frame_count) &&
               (res_buffer->channels == tgt.channels)) {
      fprintf(stderr,
              "existing file appears to be complete,"
              " can't resume processing\n");
      retval = FALSE;
      goto end;
    } else {
      fprintf(stderr,
              "existing file doesn't match target file,"
              " can't resume processing\n");
      retval = FALSE;
      goto end;
    }
  }

  /* open output file, truncating if necessary */
  outconfig = afNewFileSetup();
  afInitFileFormat(outconfig, tgt.file_format);
  afInitRate(outconfig, AF_DEFAULT_TRACK, tgt.rate);
  afInitSampleFormat(outconfig, AF_DEFAULT_TRACK, tgt.format, tgt.width);
  afInitChannels(outconfig, AF_DEFAULT_TRACK, tgt.channels);
  afInitByteOrder(outconfig, AF_DEFAULT_TRACK, tgt.byte_order);

  out.file = afOpenFD(fileno(out_stream), "w", outconfig);
  if (!out.file) {
    fprintf(stderr, "couldn't open output file as an audio file\n");
    retval = FALSE;
    goto end;
  }


  afSetVirtualByteOrder(tgt.file, AF_DEFAULT_TRACK,
                        INTERNAL_BYTEORDER);

  afSetVirtualByteOrder(out.file, AF_DEFAULT_TRACK,
                        INTERNAL_BYTEORDER);

  afSetVirtualSampleFormat(tgt.file, AF_DEFAULT_TRACK, INTERNAL_FORMAT,
                           INTERNAL_WIDTH);

  afSetVirtualSampleFormat(out.file, AF_DEFAULT_TRACK, INTERNAL_FORMAT,
                           INTERNAL_WIDTH);

  tgt.format = INTERNAL_FORMAT;
  tgt.width = INTERNAL_WIDTH;
  out.format = INTERNAL_FORMAT;
  out.width = INTERNAL_WIDTH;

  iobuf = (int16_t *)malloc(sizeof(int16_t) * tgt.channels * tile_size);
  if (iobuf == NULL) {
    fprintf(stderr, "couldn't allocate an internal buffer\n");
    retval = FALSE;
    goto end;
  }

  prebuf = (int16_t *)malloc(sizeof(int16_t) * tgt.channels * tile_size);
  if (prebuf == NULL) {
    fprintf(stderr, "couldn't allocate an internal buffer\n");
    retval = FALSE;
    goto end;
  }

  postbuf = (int16_t *)malloc(sizeof(int16_t) * tgt.channels * tile_size);
  if (postbuf == NULL) {
    fprintf(stderr, "couldn't allocate an internal buffer\n");
    retval = FALSE;
    goto end;
  }


  tick_frames = tgt.frame_count / 77;
  last_mark = 0;
  total_frames_read = 0;


  if (res_ok) {
    /* if we have a resume buffer, dump it into the output file and
       free that memory before we eat a whole bunch more memory
       buffering the source file */

    /* copy previous output data into the new output file */
    status = afWriteFrames(out.file, AF_DEFAULT_TRACK,
                           res_buffer->data, res_buffer->frame_count);
    if (status != res_buffer->frame_count) {
      fprintf(stderr, "failed writing existing data to resumed output file\n");
      retval = FALSE;
      goto end;
    }

    /* seek to the end to the corresponding point in the target file. */
    afSeekFrame(tgt.file, AF_DEFAULT_TRACK, res_buffer->frame_count);

    out.frame_count = afGetFrameCount(out.file, AF_DEFAULT_TRACK);
    assert(out.frame_count == res_buffer->frame_count);

    total_frames_read += status;

    /* sync the headers so that the data will be preserved even if the
       soundmosaic process fails immediately */
    status = afSyncFile(out.file);
    if (status != 0) {
      fprintf(stderr, "failed syncing resumed output file audio headers\n");
      retval = FALSE;
      goto end;
    }

    /* free things up */
    free_resume_buffer(res_buffer);
  }

  status = buffer_source_file(&src_buffer, src_stream, &tgt);
  if (!status) {
    fprintf(stderr, "failed buffering source file\n");
    retval = FALSE;
    goto end;
  }

  if (graphics) {
    graphics_init();
    draw = graphics_draw_match;
  }

  sm = soundmosaic_init(src_buffer.data,
                        src_buffer.len,
                        src_buffer.channels,
                        draw);

  if (dying) {
    retval = FALSE;
    goto end;
  }

  start_time = time(NULL);
  printf("performing soundmosaic\n");
  printf("[%77c]\n", ' ');
  printf(" ");

  if (res_ok) {
    /* draw tick marks to indicate the amount of processing that has
       already been done */
    while (total_frames_read - last_mark >= tick_frames) {
      last_mark += tick_frames;
      printf("#");
      fflush(stdout);
    }
  }


  do {
    frames_read = afReadFrames(tgt.file, AF_DEFAULT_TRACK, iobuf,
                               tile_size);
    assert(frames_read <= tile_size);
    //fprintf(stderr, "read %d of %d expected\n",
    //        frames_read, tile_size);

    if (frames_read > 0) {
      deinterlace(iobuf, prebuf, frames_read * tgt.channels, tgt.channels);

      for (i = 0; i < tgt.channels; i++) {
        ibuf = &prebuf[i * frames_read];
        obuf = &postbuf[i * frames_read];

        status = soundmosaic(sm, ibuf, obuf, frames_read, partition);
        if (status != SOUNDMOSAIC_SUCCESS) {
          print_soundmosaic_error(status);
          retval = FALSE;
          goto end;
        }
      }

      interlace(postbuf, iobuf, frames_read * tgt.channels, tgt.channels);

      status = afWriteFrames(out.file, AF_DEFAULT_TRACK, iobuf, frames_read);
      if (status != frames_read) {
        fprintf(stderr, "failed writing file, wrote %d of %d frames\n",
                status, frames_read);
        retval = FALSE;
        goto end;
      }

      status = afSyncFile(out.file);
      if (status != 0) {
        fprintf(stderr, "failed syncing output file audio headers\n");
        retval = FALSE;
        goto end;
      }

      total_frames_read += frames_read;
      while (total_frames_read - last_mark >= tick_frames) {
        last_mark += tick_frames;
        printf("#");
        fflush(stdout);
      }
    }
  } while((!dying) && (frames_read == tile_size));
  printf("\n");

  end_time = time(NULL);

  afSeekFrame(tgt.file, AF_DEFAULT_TRACK, 0);

  retval = TRUE;

 end:
  if (graphics && (draw != NULL)) {
    graphics_close();
  }

  soundmosaic_free(sm);
  free_source_buffer(&src_buffer);

  if (iobuf != NULL) {
    free(iobuf);
  }

  if (prebuf != NULL) {
    free(prebuf);
  }

  if (postbuf != NULL) {
    free(postbuf);
  }

  if ((out.file != NULL) && (afSyncFile(out.file) != 0)) {
    fprintf(stderr, "failed syncing output file audio headers\n");
    retval = FALSE;
  }

  if (outconfig != NULL) {
    afFreeFileSetup(outconfig);
  }

  if (retval == TRUE) {
    printf("soundmosaic finished processing source file in %d seconds\n",
           end_time - start_time);
  }
  return retval;
}


int soundmosaic_collate_streams(FILE *tgt_stream,
                                FILE *src1_stream,
                                FILE *src2_stream,
                                FILE *out_stream,
                                float tile_seconds)
{
  int tile_size;
  audio_file_info tgt;
  audio_file_info out;
  AFfilehandle src1_file = NULL;
  AFfilehandle src2_file = NULL;
  int src1_frame_total;
  int src2_frame_total;
  AFfilesetup outconfig = NULL;
  char format_name[32];
  int16_t *ibuf = NULL;
  int16_t *s1ibuf = NULL;
  int16_t *s2ibuf = NULL;
  int16_t *obuf = NULL;
  int16_t *iobuf = NULL;
  int16_t *s1iobuf = NULL;
  int16_t *s2iobuf = NULL;
  int16_t *prebuf = NULL;
  int16_t *s1prebuf = NULL;
  int16_t *s2prebuf = NULL;
  int16_t *postbuf = NULL;
  int target_frames_read, frames_read;
  int i;
  int status;
  int retval = FALSE;

  tgt.file = NULL;
  out.file = NULL;

  /* open the target file */
  tgt.file = afOpenFD(fileno(tgt_stream), "r", NULL);
  if (!tgt.file) {
    fprintf(stderr, "couldn't open target file as an audio file\n");
    retval = FALSE;
    goto end;
  }

  /* get target file parameters */
  tgt.file_format = afGetFileFormat(tgt.file, NULL);
  tgt.frame_count = afGetFrameCount(tgt.file, AF_DEFAULT_TRACK);
  tgt.channels = afGetChannels(tgt.file, AF_DEFAULT_TRACK);
  tgt.rate = afGetRate(tgt.file, AF_DEFAULT_TRACK);
  tgt.byte_order = afGetByteOrder(tgt.file, AF_DEFAULT_TRACK);
  tgt.bytes = afGetTrackBytes(tgt.file, AF_DEFAULT_TRACK);
  afGetSampleFormat(tgt.file, AF_DEFAULT_TRACK, &tgt.format, &tgt.width);

  tile_size = tgt.rate * tile_seconds;
  assert(tile_size >= 0);
  if (tile_size == 0) {
    fprintf(stderr, "tile size too small: less than one sample per frame\n");
    retval = FALSE;
    goto end;
  }

  switch (tgt.format) {
  case AF_SAMPFMT_DOUBLE:
    strcpy(format_name, "DOUBLE");
    break;
  case AF_SAMPFMT_UNSIGNED:
    strcpy(format_name, "UNSIGNED");
    break;
  case AF_SAMPFMT_TWOSCOMP:
    strcpy(format_name, "TWOSCOMP");
    break;
  case AF_SAMPFMT_FLOAT:
    strcpy(format_name, "FLOAT");
    break;
  default:
    strcpy(format_name, "UNKNOWN");
  }

  //fprintf(stderr, "target file:\n");

  //fprintf(stderr,
  //      "frames: %d, channels: %d, rate: %d, format: %s, width: %d\n",
  //      tgt.frame_count, tgt.channels, tgt.rate, format_name, tgt.width);

  //fprintf(stderr, "tile size: %d frames, %f seconds\n", tile_size,
  //        ((float)tile_size / (float)tgt.rate));



  /* create and open output file */
  outconfig = afNewFileSetup();
  afInitFileFormat(outconfig, tgt.file_format);
  afInitRate(outconfig, AF_DEFAULT_TRACK, tgt.rate);
  afInitSampleFormat(outconfig, AF_DEFAULT_TRACK, tgt.format, tgt.width);
  afInitChannels(outconfig, AF_DEFAULT_TRACK, tgt.channels);
  afInitByteOrder(outconfig, AF_DEFAULT_TRACK, tgt.byte_order);

  out.file = afOpenFD(fileno(out_stream), "w", outconfig);
  if (!out.file) {
    fprintf(stderr, "couldn't open output file as an audio file\n");
    retval = FALSE;
    goto end;
  }

  afSetVirtualByteOrder(tgt.file, AF_DEFAULT_TRACK,
                        INTERNAL_BYTEORDER);

  afSetVirtualByteOrder(out.file, AF_DEFAULT_TRACK,
                        INTERNAL_BYTEORDER);

  afSetVirtualSampleFormat(tgt.file, AF_DEFAULT_TRACK, INTERNAL_FORMAT,
                           INTERNAL_WIDTH);

  afSetVirtualSampleFormat(out.file, AF_DEFAULT_TRACK, INTERNAL_FORMAT,
                           INTERNAL_WIDTH);

  tgt.format = INTERNAL_FORMAT;
  tgt.width = INTERNAL_WIDTH;
  out.format = INTERNAL_FORMAT;
  out.width = INTERNAL_WIDTH;

  /* open attempt file 1 */
  src1_file = afOpenFD(fileno(src1_stream), "r", NULL);
  if (!src1_file) {
    fprintf(stderr, "couldn't open an attempt file as an audio file\n");
    retval = FALSE;
    goto end;
  }

  src1_frame_total = afGetFrameCount(src1_file, AF_DEFAULT_TRACK);

  if (src1_frame_total != tgt.frame_count ||
      afGetChannels(src1_file, AF_DEFAULT_TRACK) != tgt.channels ||
      afGetRate(src1_file, AF_DEFAULT_TRACK) != tgt.rate ||
      afGetTrackBytes(src1_file, AF_DEFAULT_TRACK) != tgt.bytes) {
    fprintf(stderr,
            "an attempt file format doesn't match target file format\n");
    retval = FALSE;
    goto end;
  }

  afSetVirtualByteOrder(src1_file, AF_DEFAULT_TRACK,
                        INTERNAL_BYTEORDER);

  afSetVirtualSampleFormat(src1_file, AF_DEFAULT_TRACK, INTERNAL_FORMAT,
                           INTERNAL_WIDTH);


  /* open attempt file 2 */
  src2_file = afOpenFD(fileno(src2_stream), "r", NULL);
  if (!src2_file) {
    fprintf(stderr, "couldn't open an attempt file as an audio file\n");
    retval = FALSE;
    goto end;
  }

  src2_frame_total = afGetFrameCount(src2_file, AF_DEFAULT_TRACK);

  if (src2_frame_total != tgt.frame_count ||
      afGetChannels(src2_file, AF_DEFAULT_TRACK) != tgt.channels ||
      afGetRate(src2_file, AF_DEFAULT_TRACK) != tgt.rate ||
      afGetTrackBytes(src2_file, AF_DEFAULT_TRACK) != tgt.bytes) {
    fprintf(stderr,
            "an attempt file format doesn't match target file format\n");
    retval = FALSE;
    goto end;
  }

  afSetVirtualByteOrder(src2_file, AF_DEFAULT_TRACK,
                        INTERNAL_BYTEORDER);

  afSetVirtualSampleFormat(src2_file, AF_DEFAULT_TRACK, INTERNAL_FORMAT,
                           INTERNAL_WIDTH);

  iobuf = (int16_t *)malloc(sizeof(int16_t) * tgt.channels * tile_size);
  if (iobuf == NULL) {
    fprintf(stderr, "couldn't allocate an internal buffer\n");
    retval = FALSE;
    goto end;
  }

  s1iobuf = (int16_t *)malloc(sizeof(int16_t) * tgt.channels * tile_size);
  if (s1iobuf == NULL) {
    fprintf(stderr, "couldn't allocate an internal buffer\n");
    retval = FALSE;
    goto end;
  }

  s2iobuf = (int16_t *)malloc(sizeof(int16_t) * tgt.channels * tile_size);
  if (s2iobuf == NULL) {
    fprintf(stderr, "couldn't allocate an internal buffer\n");
    retval = FALSE;
    goto end;
  }

  prebuf = (int16_t *)malloc(sizeof(int16_t) * tgt.channels * tile_size);
  if (prebuf == NULL) {
    fprintf(stderr, "couldn't allocate an internal buffer\n");
    retval = FALSE;
    goto end;
  }

  s1prebuf = (int16_t *)malloc(sizeof(int16_t) * tgt.channels * tile_size);
  if (s1prebuf == NULL) {
    fprintf(stderr, "couldn't allocate an internal buffer\n");
    retval = FALSE;
    goto end;
  }

  s2prebuf = (int16_t *)malloc(sizeof(int16_t) * tgt.channels * tile_size);
  if (s2prebuf == NULL) {
    fprintf(stderr, "couldn't allocate an internal buffer\n");
    retval = FALSE;
    goto end;
  }

  postbuf = (int16_t *)malloc(sizeof(int16_t) * tgt.channels * tile_size);
  if (postbuf == NULL) {
    fprintf(stderr, "couldn't allocate an internal buffer\n");
    retval = FALSE;
    goto end;
  }


  do {
    target_frames_read = afReadFrames(tgt.file, AF_DEFAULT_TRACK, iobuf,
                                      tile_size);
    if (target_frames_read > 0) {

      frames_read = afReadFrames(src1_file, AF_DEFAULT_TRACK, s1iobuf,
                                 target_frames_read);
      if (frames_read != target_frames_read) {
        fprintf(stderr, "error reading attempt file, read %d of %d frames\n",
                frames_read, target_frames_read);
        retval = FALSE;
        goto end;
      }

      frames_read = afReadFrames(src2_file, AF_DEFAULT_TRACK, s2iobuf,
                                 target_frames_read);
      if (frames_read != target_frames_read) {
        fprintf(stderr, "error reading attempt file, read %d of %d frames\n",
                frames_read, target_frames_read);
        retval = FALSE;
        goto end;
      }

      deinterlace(iobuf, prebuf, frames_read * tgt.channels, tgt.channels);
      deinterlace(s1iobuf, s1prebuf, frames_read * tgt.channels, tgt.channels);
      deinterlace(s2iobuf, s2prebuf, frames_read * tgt.channels, tgt.channels);

      for (i = 0; i < tgt.channels; i++) {
        ibuf = &prebuf[i * frames_read];
        s1ibuf = &s1prebuf[i * frames_read];
        s2ibuf = &s2prebuf[i * frames_read];
        obuf = &postbuf[i * frames_read];


        status = soundmosaic_collate(ibuf, s1ibuf, s2ibuf, obuf,
                                     frames_read);
        if (status != SOUNDMOSAIC_SUCCESS) {
          print_soundmosaic_error(status);
          retval = FALSE;
          goto end;
        }
      }

      interlace(postbuf, iobuf, frames_read * tgt.channels, tgt.channels);

      status = afWriteFrames(out.file, AF_DEFAULT_TRACK, iobuf, frames_read);
      if (status != frames_read) {
        fprintf(stderr, "failed writing file, wrote %d of %d frames\n",
                status, frames_read);
        retval = FALSE;
        goto end;
      }

      status = afSyncFile(out.file);
      if (status != 0) {
        fprintf(stderr, "failed syncing output file audio headers\n");
        retval = FALSE;
        goto end;
      }
    }
  } while (target_frames_read == tile_size);

  retval = TRUE;

 end:
  free(iobuf);
  free(s1iobuf);
  free(s2iobuf);
  free(prebuf);
  free(s1prebuf);
  free(s2prebuf);
  free(postbuf);

  if ((out.file != NULL) && (afSyncFile(out.file) != 0)) {
    fprintf(stderr, "failed syncing output file audio headers\n");
    retval = FALSE;
  }

  if (outconfig != NULL) {
    afFreeFileSetup(outconfig);
  }

  printf("soundmosaic collate finished\n");
  return retval;
}


int add_src(const char *name)
{
  if (srcs_root == NULL) {
    srcs_root = (node *)malloc(sizeof(node));
    if (srcs_root == NULL) {
      perror("");
      return FALSE;
    }
    srcs_end = srcs_root;
    srcs_end->this = NULL;
    srcs_end->next = NULL;
  } else {
    assert(srcs_end != NULL);
    assert(srcs_end->next == NULL);

    srcs_end->next = (node *)malloc(sizeof(node));
    if (srcs_end->next == NULL) {
      perror("couldn't allocate memory for source node");
      return FALSE;
    }
    srcs_end = srcs_end->next;
    srcs_end->this = NULL;
    srcs_end->next = NULL;
  }
  srcs_end->this = malloc(PATH_MAX+1);
  if (srcs_end->this == NULL) {
    perror("couldn't allocate memory for source stream");
    return FALSE;
  }
  strncpy((char *)srcs_end->this, name, PATH_MAX);
  ((char*)srcs_end->this)[PATH_MAX] = 0;

  return TRUE;
}


int add_attempt(FILE *stream)
{
  if (attempts_root == NULL) {
    assert(stream == NULL);
    attempts_root = (node *)malloc(sizeof(node));
    if (attempts_root == NULL) {
      perror("");
      return FALSE;
    }
    attempts_end = attempts_root;
    attempts_end->this = NULL;
    attempts_end->next = NULL;
  } else {
    assert(attempts_end != NULL);
    assert(attempts_end->next == NULL);

    attempts_end->next = (node *)malloc(sizeof(node));
    if (attempts_end->next == NULL) {
      perror("couldn't allocate memory for attempt node");
      return FALSE;
    }
    attempts_end = attempts_end->next;
    attempts_end->this = NULL;
    attempts_end->next = NULL;

    attempts_end->this = stream;
    if (attempts_end->this == NULL) {
      fprintf(stderr, "added invalid NULL attempt stream\n");
      return FALSE;
    }
  }

  return TRUE;
}


int soundmosaic_srcs (FILE *tgt_stream,
                      FILE *out_stream,
                      float tile_seconds,
                      int partition,
                      int graphics,
                      resume_buffer *res_buffer)
{
  int status;
  int retval = TRUE;
  int i;
  node *srcs_cur = NULL;
  FILE *temp_streams[3];

  assert(tgt_stream != NULL);

  srcs_cur = srcs_root;

  for (i = 0; i < 3; i++) {
    temp_streams[i] = tmpfile();
    if (temp_streams[i] == NULL) {
      perror("error opening temp file");
      return FALSE;
    }
  }

  i = 0;
  while ((!dying) && (srcs_cur != NULL)) {
    FILE *tempsrc_stream = NULL;
    FILE *tempout_stream = NULL;

    assert(srcs_cur != NULL);
    assert(srcs_cur->this != NULL);

    if (i > 0) {
      printf("\n");
    }
    printf("processing source file %s\n", srcs_cur->this);

    if (i == 0) {
      tempout_stream = temp_streams[0];
    } else {
      tempout_stream = temp_streams[1];
    }

    tempsrc_stream = fopen(srcs_cur->this, "r");
    if (tempsrc_stream == NULL) {
      perror("couldn't open source file");
      retval = FALSE;
      break;
    }

    status = soundmosaic_streams(tgt_stream, tempsrc_stream,
                                 tempout_stream, tile_seconds,
                                 partition, graphics, res_buffer);
    if (!status) {
      fprintf(stderr, "soundmosaic failed\n");
      retval = FALSE;
      break;
    }

    fseek(tgt_stream, 0, SEEK_SET);
    fseek(tempout_stream, 0, SEEK_SET);

    fclose(tempsrc_stream);

    if ((i > 0) && (!dying)) {
      printf("collating\n");
      status = soundmosaic_collate_streams(tgt_stream,
                                           temp_streams[0],  // src 1
                                           temp_streams[1],  // src 2
                                           temp_streams[2],  // output
                                           tile_seconds);
      if (!status) {
        fprintf(stderr, "soundmosaic collate failed\n");
        retval = FALSE;
        break;
      }

      fseek(tgt_stream, 0, SEEK_SET);

      // close 0, put 2 in slot 0, open a new 2
      fclose(temp_streams[0]);
      temp_streams[0] = temp_streams[2];
      fseek(temp_streams[0], 0, SEEK_SET);
      temp_streams[2] = tmpfile();
      if (temp_streams[2] == NULL) {
        perror("error opening temp file");
        retval = FALSE;
        break;
      }

      // close 1 and start it over
      fclose(temp_streams[1]);
      temp_streams[1] = tmpfile();
      if (temp_streams[1] == NULL) {
        perror("error opening temp file");
        retval = FALSE;
        break;
      }

      tempout_stream = temp_streams[1];
    }

    srcs_cur = srcs_cur->next;
    i++;
  }

  printf("writing output file\n");
  fseek(temp_streams[0], 0, SEEK_SET);
  status = copy_stream(temp_streams[0], out_stream);
  if (!status) {
    fprintf(stderr, "failed copying temporary file to output file\n");
    retval = FALSE;
  }
  fclose(temp_streams[0]);

  return retval;
}


int soundmosaic_collate_attempts(FILE *tgt_stream,
                                 FILE *out_stream,
                                 int master,
                                 float tile_seconds)
{
  int status;
  int i;
  int retval = TRUE;
  node *attempts_cur = NULL;
  FILE *temp_streams[3];

  attempts_cur = attempts_root;

  // the root node for attempts is always NULL, since we might start
  // by waiting for attempt files from the slaves
  assert(attempts_cur->this == NULL);

  i = 0;
  while (!dying && (master || attempts_cur->next != NULL)) {

    if (attempts_cur->next == NULL && master) {
      assert(master);
      printf("\nwaiting for attempt files from slaves,"
             " press Ctrl-C to finish\n");
      while ((attempts_cur->next == NULL) && (!dying)) {
        sleep(5);
      }
    }


    if (!dying && attempts_cur->next != NULL) {
      attempts_cur = attempts_cur->next;
      assert(attempts_cur != NULL);
      assert(attempts_cur->this != NULL);

      if (i == 0) {
        temp_streams[0] = attempts_cur->this;
      } else {
        temp_streams[1] = attempts_cur->this;
      }

      if (i > 0) {
        FILE *tempout_stream = NULL;

        tempout_stream = tmpfile();
        if (tempout_stream == NULL) {
          perror("error opening temp file");
          retval = FALSE;
          break;
        }

        printf("collating\n");
        status = soundmosaic_collate_streams(tgt_stream,
                                             temp_streams[0],
                                             temp_streams[1],
                                             tempout_stream,
                                             tile_seconds);
        if (!status) {
          fprintf(stderr, "soundmosaic collate failed\n");
          retval = FALSE;
          break;
        }

        fseek(tgt_stream, 0, SEEK_SET);

        // close 0, put tempout in slot 0
        fclose(temp_streams[0]);
        temp_streams[0] = tempout_stream;
        fseek(temp_streams[0], 0, SEEK_SET);

        // close 1 so we can start it over with the next file
        fclose(temp_streams[1]);
      }
    }

    i++;
  }

  if (i > 0) {
    printf("writing output file\n");
    fseek(temp_streams[0], 0, SEEK_SET);
    status = copy_stream(temp_streams[0], out_stream);
    if (!status) {
      fprintf(stderr, "failed copying temporary file to output file\n");
      retval = FALSE;
    }
    fclose(temp_streams[0]);
  }

  return retval;
}


int main(int argc, char **argv)
{
  int status = -1;
  pthread_t server_thread;

  char tgt_name[PATH_MAX+1] = {0};
  char out_name[PATH_MAX+1] = {0};
  char url[PATH_MAX+1];
  char *hostname = NULL;
  char *portstr;
  char *pathstart;

  float tile_seconds = 0.1;
  int num_srcs = 0;
  int num_attempts = 0;
  int collate = FALSE;
  int resume = FALSE;
  int master = FALSE;
  int slave = FALSE;
  int partition = FALSE;
  int graphics = FALSE;

  int port = 0;

  FILE *tgt_stream = NULL;
  FILE *out_stream = NULL;

  node *srcs_cur = NULL;
  node *attempts_cur = NULL;

  resume_buffer *res_buffer = NULL;

  server_args_type server_args;

  int c;
  int option_index = 0;
  static struct option long_options[] =
  {
    /* name, argument?, NULL, short name */
    {"tilesize", 1, NULL, 't'},
    {"collate", 0, NULL, 'c'},
    {"resume", 0, NULL, 'r'},
    {"partition", 0, NULL, 'p'},
    {"master", 1, NULL, 'm'},
    {"slave", 1, NULL, 's'},
    {"graphics", 0, NULL, 'g'},
    {0, 0, 0, 0}
  };


  while ((c = getopt_long (argc, argv, "t:s:m:crpg",
                           long_options, &option_index)) != -1) {

    switch (c) {
    case 't':
      sscanf(optarg, "%f", &tile_seconds);
      break;
    case 's':
      slave = TRUE;
      strncpy(url, optarg, PATH_MAX);
      url[PATH_MAX] = 0;
      break;
    case 'm':
      master = TRUE;
      collate = TRUE;
      port = atoi(optarg);
      break;
    case 'c':
      collate = TRUE;
      break;
    case 'r':
      resume = TRUE;
      break;
    case 'p':
      partition = TRUE;
      break;
    case 'g':
      graphics = TRUE;
      break;

    case 'h':
    case '?':
      usage();
      return 0;

    default:
      printf("An option (-%c) is defined but not actually handled\n", c);
      printf("This is a bug -- please let the maintainer know.\n");
      usage();
      return 1;
    }

  }

  // if we're not a slave, we'll have target and output file names on
  // the command line
  if (!slave) {
    int file_length;

    if (argc - optind < 2) {
      usage();
      return 1;
    }

    strncpy(tgt_name, argv[optind], PATH_MAX);
    tgt_name[PATH_MAX] = 0;
    tgt_stream = fopen(tgt_name, "r");
    if (tgt_stream == NULL) {
      perror("couldn't open target file");
      return 1;
    }
    optind++;

    strncpy(out_name, argv[optind], PATH_MAX);
    out_name[PATH_MAX] = 0;
    out_stream = fopen(out_name, "a+");
    if (out_stream == NULL) {
      perror("couldn't open output file");
      return 1;
    }
    optind++;

    fseek(out_stream, 0, SEEK_END);
    file_length = ftell(out_stream);
    fseek(out_stream, 0, SEEK_SET);

    if (file_length > 0) {
      if (resume) {
        res_buffer = malloc(sizeof(resume_buffer));
        if (res_buffer == NULL) {
          fprintf(stderr,
                  "failed to allocate %d bytes for a resume file buffer\n",
                  sizeof(resume_buffer));
          return 1;
        }

        status = buffer_resume_stream(out_stream, res_buffer);
        if (status != TRUE) {
          fprintf(stderr, "failed reading existing output file\n");
          return 1;
        }

        fclose(out_stream);

        /* reopen in read/write mode.  This will position the stream
           at the beginning of the file, so we should overwrite the
           old data.  But if something goes wrong before we can do
           that, the old data won't be erased. */
        out_stream = fopen(out_name, "r+");
        if (out_stream == NULL) {
          perror("couldn't open output file");
          return 1;
        }
      } else {
        fprintf(stderr, "output file already exists\n");
        return 1;
      }
    }
  }

  // if we are a slave, we need use temp files for the target and
  // output, and ask the server for the target data
  if (slave) {
    tgt_stream = tmpfile();
    if (tgt_stream == NULL) {
      perror("couldn't open temp file for target data\n");
      return 1;
    }

    out_stream = tmpfile();
    if (out_stream == NULL) {
      perror("couldn't open temp file for output data\n");
      return 1;
    }

    if (strncasecmp(url, "http://", 7) != 0) {
      fprintf(stderr, "invalid URL '%s'\n", url);
      usage();
      return 1;
    }

    hostname = &url[7];

    pathstart = strchr(hostname, '/');
    if (pathstart == NULL) {
      fprintf(stderr, "invalid URL '%s'\n", url);
      usage();
      return 1;
    }
    pathstart[0] = 0;

    portstr = strchr(hostname, ':');
    if (portstr == NULL) {
      port = 80;
    } else {
      portstr[0] = 0;
      portstr++;

      port = atoi(portstr);
    }

    printf("requesting target file from master at %s:%d\n", hostname, port);

    status = get_target_file(hostname, port, tgt_stream);
    if (!status) {
      fprintf(stderr, "failed getting target file from master\n");
      return 1;
    }
  }

  // now read source or attempt file names from the command line
  if (collate) {
    // for collation, we always use a null attempt_root, since no files
    // have to be specified.
    add_attempt(NULL);
    while (optind < argc) {
      FILE *attempt_stream;
      attempt_stream = fopen(argv[optind], "r");
      if (attempt_stream == NULL) {
        perror( "couldn't open an attempt file" );
        return 1;
      }
      status = add_attempt(attempt_stream);
      if (!status) {
        return status;
      }
      optind++;
      num_attempts++;
    }
  } else {
    while (optind < argc) {
      status = add_src(argv[optind]);
      if (!status) {
        return status;
      }
      optind++;
      num_srcs++;
    }
  }

  if (signal(SIGINT, termination_handler) == SIG_IGN) {
    signal(SIGINT, SIG_IGN);
  }
  if (signal(SIGTERM, termination_handler) == SIG_IGN) {
    signal(SIGTERM, SIG_IGN);
  }


  if (collate) {

    if (resume || slave || (!master && num_attempts < 2)) {
      usage();
      return 1;
    }

    if (master) {
      server_args.port = port;
      server_args.tgt_name = tgt_name;
      status = pthread_create(&server_thread, NULL, serve, &server_args);
      if (status != 0) {
        fprintf(stderr, "error creating server thread: %d\n", status);
        return 1;
      }
    }

    assert(tgt_stream != NULL);

    status = soundmosaic_collate_attempts(tgt_stream,
                                          out_stream,
                                          master,
                                          tile_seconds);

  } else {

    if (master || (num_srcs == 0) || (resume == TRUE && num_srcs != 1)) {
      usage();
      return 1;
    }

    if (resume) {
      assert(res_buffer != NULL);
    } else {
      assert(res_buffer == NULL);
    }

    status = soundmosaic_srcs(tgt_stream,
                              out_stream,
                              tile_seconds,
                              partition,
                              graphics,
                              res_buffer);

    if (slave && status && (!dying)) {
      printf("sending output to master\n");
      send_output_file(hostname, port, out_stream);
    }
  }

  // free srcs
  srcs_cur = srcs_root;
  while (srcs_cur != NULL) {
    node *dead_node;

    if (srcs_cur->this != NULL) {
      free(srcs_cur->this);
    }

    dead_node = srcs_cur;
    srcs_cur = srcs_cur->next;
    free(dead_node);
  }

  // free attempts
  attempts_cur = attempts_root;
  while (attempts_cur != NULL) {
    node *dead_node;

    dead_node = attempts_cur;
    attempts_cur = attempts_cur->next;
    free(dead_node);
  }

  fclose(tgt_stream);
  fclose(out_stream);

  if (dying && !master) {
    fprintf(stderr, "Killed\n");
  }

  if (status == TRUE) {
    return 0;
  }

  return 1;
}
