/* soundmosaic, Copyright (c) 2001-2007 Steven Hazel <sah@thalassocracy.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */
#include "config.h"

#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <time.h>
#include <assert.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "main.h"
#include "io_buffer.h"
#include "http.h"
#include "server.h"

#define CONN_NOT_CONNECTED 1
#define CONN_CONNECTED 2
#define CONN_PENDING_CLOSE 3
#define CONN_CLOSED 4

#define FILE_NOT_STARTED 1
#define FILE_SENDING 2
#define FILE_RECEIVING 3

typedef struct {
  int sock;
  int state;
  char addr[16];
  int port;
  int file_state;
  FILE *file;
  int file_pos;
  int file_size;
  time_t last_read_time;
  time_t last_write_time;
  io_buffer in_buf;
  io_buffer out_buf;
} connection;

extern int dying;

server_args_type *args;

node *connections_root = NULL;
node *connections_end = NULL;


int add_connection(connection *conn)
{
  if (connections_root == NULL) {
    connections_root = (node *)malloc(sizeof(node));
    if (connections_root == NULL) {
      perror("");
      return FALSE;
    }
    connections_end = connections_root;
    connections_end->this = NULL;
    connections_end->next = NULL;
  } else {
    assert(connections_end != NULL);
    assert(connections_end->next == NULL);

    connections_end->next = (node *)malloc(sizeof(node));
    if (connections_end->next == NULL) {
      perror("");
      return FALSE;
    }
    connections_end = connections_end->next;
    connections_end->this = NULL;
    connections_end->next = NULL;
  }

  connections_end->this = conn;
  if (connections_end->this == NULL) {
    perror("");
    return FALSE;
  }

  return TRUE;
}


void init_connection(connection *conn)
{
  conn->state = CONN_NOT_CONNECTED;
  conn->file = NULL;
  conn->file_state = FILE_NOT_STARTED;

  io_buffer_init(&conn->in_buf, 64 * 1024);
  io_buffer_init(&conn->out_buf, 64 * 1024);

  conn->port = args->port;
  strcpy(conn->addr, "127.0.0.1");

  add_connection(conn);
}


int close_connection(connection *conn)
{
  int status;

  assert(conn->state != CONN_NOT_CONNECTED);
  assert(conn->state != CONN_CLOSED);

  io_buffer_destroy(&conn->in_buf);
  io_buffer_destroy(&conn->out_buf);

  status = close(conn->sock);

  conn->state = CONN_CLOSED;
  conn->sock = -1;

  return status;
}


int set_connection_send_file(connection *conn, char *filename)
{
  int status;
  struct stat info;

  conn->file = fopen(filename, "r");
  if (conn->file == NULL) {
    perror("failed to open file to send");
    return FALSE;
  }

  status = fstat(fileno(conn->file), &info);
  if (status != 0) {
    perror("failed to stat file to send");
    return FALSE;
  }

  conn->file_size = info.st_size;
  conn->file_pos = 0;

  return TRUE;
}


int set_connection_recv_file(connection *conn)
{
  conn->file = tmpfile();
  if (conn->file == NULL) {
    perror("error opening temp file\n");
    close_connection(conn);
    return FALSE;
  }
  conn->file_size = 0;
  conn->file_pos = 0;
  conn->last_read_time = -1;
  conn->last_write_time = -1;

  return TRUE;
}


void handle_connection(connection *conn)
{
  http_request request;
  const char *common_headers =
    "HTTP/1.0 200 OK\x0D\x0A"
    "Server: soundmosaic/" VERSION "\x0D\x0A"
    "Connection: Close\x0D\x0A";
  const char *content_type_header =
    "Content-Type: audio/unknown\x0D\x0A";
  const char *end_headers = "\x0D\x0A";
  char content_length_header[80];
  int content_length = -1;
  int count;
  int status;
  int i;

  if (conn->state == CONN_PENDING_CLOSE) {
    if (io_buffer_bytes_used(&conn->out_buf) == 0) {
      close_connection(conn);
    }
    return;
  }

  assert(conn->state == CONN_CONNECTED);

  if (conn->file_state == FILE_NOT_STARTED) {
    char *headers;
    int header_end_pos =
      http_find_header_end(&conn->in_buf, 1);

    if (header_end_pos >= 0) {
      headers = malloc(header_end_pos + 1);
      if (headers == NULL) {
        fprintf(stderr, "couldn't allocate %d bytes for HTTP headers\n",
                header_end_pos + 1);
        return;
      }

      io_buffer_read(&conn->in_buf,
                     headers,
                     header_end_pos);
      headers[header_end_pos] = 0;

      http_parse_headers(headers, &request);

      free(headers);

      for (i = 0; i < request.num_headers; i++) {
        if (strcasecmp(request.keys[i], "content-length") == 0) {
          content_length = atoi(request.values[i]);
        }
      }
      //fprintf(stderr, "got request type %d, content length %d\n",
      //        request.type, content_length);

      if (request.type == HTTP_GET) {
        printf("sending target file to slave\n");
        status = set_connection_send_file(conn, args->tgt_name);
        if (!status) {
          fprintf(stderr, "error setting file");
          close_connection(conn);
          return;
        }

        io_buffer_write(&conn->out_buf,
                        common_headers,
                        strlen(common_headers));

        io_buffer_write(&conn->out_buf,
                        content_type_header,
                        strlen(content_type_header));

        snprintf(content_length_header, 80,
                 "Content-Length: %d\x0D\x0A", conn->file_size);

        io_buffer_write(&conn->out_buf,
                        content_length_header,
                        strlen(content_length_header));

        io_buffer_write(&conn->out_buf,
                        end_headers,
                        strlen(end_headers));

        conn->file_state = FILE_SENDING;
      } else if (request.type == HTTP_POST) {
        printf("receiving output attempt file from slave\n");
        status = set_connection_recv_file(conn);
        if (!status) {
          fprintf(stderr, "error setting file");
          close_connection(conn);
          return;
        }

        if (content_length < 1) {
          fprintf(stderr, "invalid Content-Length %d from slave\n",
                  content_length);
          close_connection(conn);
          return;
        }

        conn->file_size = content_length;
        conn->file_state = FILE_RECEIVING;
      } else {
        fprintf(stderr, "unknown HTTP command, dropping connection\n");
        close_connection(conn);
        return;
      }
    }
  } else if (conn->file_state == FILE_SENDING) {
    if (io_buffer_bytes_free(&conn->out_buf) > 0) {
      status = io_buffer_fread_all(&conn->out_buf, conn->file);
      if (status > 0) {
        conn->file_pos = ftell(conn->file);
      } else {
        //fprintf(stderr, "read %d on http sending file, errno %d\n",
        //        status, errno);

        printf("finished sending target file to slave\n");
        fclose(conn->file);

        conn->file_state = FILE_NOT_STARTED;
        //fprintf(stderr, "closing connection, out buffer has %d bytes\n",
        //        io_buffer_bytes_used(&conn->out_buf));

        io_buffer_write(&conn->out_buf,
                        end_headers,
                        strlen(end_headers));

        conn->state = CONN_PENDING_CLOSE;
      }
    }
  } else if (conn->file_state == FILE_RECEIVING) {
    count = io_buffer_bytes_used(&conn->in_buf);
    assert(count > 0);
    if (count > 0) {
      if (conn->file_pos + count > conn->file_size) {
        count = conn->file_size - conn->file_pos;
      }
      status = io_buffer_fwrite(&conn->in_buf, conn->file, count);
      if (status != count) {
        if (status < 0) {
          perror("error writing file");
        }
        fprintf(stderr, "failed writing file, closing connection\n");
        close_connection(conn);
        return;
      }

      conn->file_pos = ftell(conn->file);

      assert(conn->file_pos <= conn->file_size);

      if (conn->file_pos == conn->file_size) {
        printf("finished receiving an output attempt file from slave\n");
        status = add_attempt(conn->file);
        if (!status) {
          fprintf(stderr,
                  "failed adding file as attempt, closing connection\n");
          close_connection(conn);
          return;
        }

        conn->file_state = FILE_NOT_STARTED;
        //fprintf(stderr, "closing connection, in buffer has %d bytes\n",
        //        io_buffer_bytes_used(&conn->in_buf));

        io_buffer_write(&conn->out_buf,
                        common_headers,
                        strlen(common_headers));

        io_buffer_write(&conn->out_buf,
                        "Content-type: text/plain\x0D\x0A",
                        strlen("Content-type: text/plain\x0D\x0A"));

        snprintf(content_length_header, 80,
                 "Content-Length: %d\x0D\x0A", 2);

        io_buffer_write(&conn->out_buf,
                        content_length_header,
                        strlen(content_length_header));

        io_buffer_write(&conn->out_buf,
                        end_headers,
                        strlen(end_headers));

        io_buffer_write(&conn->out_buf,
                        "OK",
                        2);

        io_buffer_write(&conn->out_buf,
                        end_headers,
                        strlen(end_headers));

        conn->state = CONN_PENDING_CLOSE;
      }
    }
  }

  if (io_buffer_failed(&conn->in_buf) ||
      io_buffer_failed(&conn->out_buf)) {
    close_connection(conn);
    return;
  }
}


void set_fdsets(connection *conn,
                unsigned int *max_fd,
                fd_set *rset,
                fd_set *wset,
                fd_set *eset) {
  int sets_changed = FALSE;

  if (conn->sock == -1) {
    return;
  }

  assert(conn->state != CONN_NOT_CONNECTED);
  assert(conn->state != CONN_CLOSED);

  if (!io_buffer_is_empty(&conn->out_buf)) {
    FD_SET((unsigned int)conn->sock, wset);
    sets_changed = TRUE;
  }

  if (!io_buffer_is_full(&conn->in_buf)) {
    FD_SET((unsigned int)conn->sock, rset);
    sets_changed = TRUE;
  }

  if (sets_changed && conn->sock > *max_fd) {
    *max_fd = conn->sock;
  }
}


void handle_read_socket_set(connection *conn) {
  int result;

  conn->last_read_time = time(NULL);

  assert(conn->state != CONN_NOT_CONNECTED);
  assert(conn->state != CONN_CLOSED);

  result = io_buffer_recv_all(&conn->in_buf, conn->sock);

  if (result == 0) {
    printf("read 0 from conn to %s:%d, state: %d\n",
           conn->addr, conn->port, conn->state);
    close_connection(conn);
  } else if (result == -1) {
    if (errno == EAGAIN) {
      // this isn't a failure
      return;
    }

    printf("read failed with err %d for conn to %s:%d, state: %d\n",
           errno, conn->addr, conn->port, conn->state);
    close_connection(conn);
  }
}


void handle_write_socket_set(connection *conn) {
  int result;

  conn->last_write_time = time(NULL);

  assert(conn->state != CONN_NOT_CONNECTED);
  assert(conn->state != CONN_CLOSED);

  result = io_buffer_send_all(&conn->out_buf, conn->sock);

  if (result == 0) {
    printf("write 0 from conn to %s:%d, state: %d\n",
           conn->addr, conn->port, conn->state);
    close_connection(conn);
  } else if (result == -1) {
    if (errno == EAGAIN) {
      // this isn't a failure
      return;
    }

    printf("write failed with err %d for conn to %s:%d, state: %d\n",
           errno, conn->addr, conn->port, conn->state);
    close_connection(conn);
  }
}


void handle_socket_sets(connection *conn,
                        fd_set *rset, fd_set *wset, fd_set *eset) {

  if (conn->sock != -1) {
    if (conn->sock != -1 &&
        FD_ISSET(conn->sock, rset)) {
      conn->last_read_time = time(NULL);
      handle_read_socket_set(conn);
    }

    if (conn->sock != -1 &&
        FD_ISSET(conn->sock, wset)) {
      conn->last_write_time = time(NULL);
      handle_write_socket_set(conn);
    }
  }
}


void free_closed_connections()
{
  node *cur_node;
  node *prev_node;
  node *dead_node;
  connection *conn;

  prev_node = NULL;
  cur_node = connections_root;
  while (cur_node != NULL) {
    assert(cur_node->this != NULL);
    conn = (connection *)cur_node->this;
    if (conn->state == CONN_CLOSED) {
      free(conn);
      if (prev_node == NULL) {
        connections_root = cur_node->next;
      } else {
        prev_node->next = cur_node->next;
      }

      dead_node = cur_node;
      cur_node = cur_node->next;
      free(dead_node);
    } else {
      prev_node = cur_node;
      cur_node = cur_node->next;
    }
  }

  cur_node = connections_root;
  while (cur_node != NULL) {
    connections_end = cur_node;
    cur_node = cur_node->next;
  }
}


void *serve(void *data)
{
  struct sockaddr_in stream_addy;
  int stream_lsocket;
  int reuse_addr = 1;
  unsigned int b;
  int flags;
  int connected_socket = -1;
  struct sockaddr_in incoming;
  time_t starttime;
  time_t start_time = 0;
  fd_set r_set, w_set, e_set;
  unsigned int max_fd;
  int num_fd;
  struct timeval timeout;
  int status;
  node *cur_node;
  connection *conn;

  args = (server_args_type *)data;

  printf("listening on port %d, target file is %s\n",
         args->port, args->tgt_name);

  start_time = time(NULL);

  /* set up a socket to listen for stream requests */

  /* Setup internet stream_addy information.
     This is used with the bind() call */
  memset((char *) &stream_addy, 0, sizeof(stream_addy));
  stream_addy.sin_family = AF_INET;
  stream_addy.sin_port = htons(args->port);
  stream_addy.sin_addr.s_addr = htonl(INADDR_ANY);

  stream_lsocket = socket(AF_INET, SOCK_STREAM, 0);
  if (stream_lsocket < 0) {
    perror("failed creating listen socket");
    goto end;
  }

  setsockopt(stream_lsocket, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse_addr,
             sizeof(reuse_addr));

  if (bind(stream_lsocket, (struct sockaddr *)&stream_addy,
           sizeof(stream_addy)) < 0) {
    perror("failed binding listen socket");
    close(stream_lsocket);
    goto end;
  }

  if (listen(stream_lsocket, 1) < 0) {
    perror("couldn't listen on socket");
    goto end;
  }

  flags = fcntl(stream_lsocket, F_GETFL);
  if (flags == -1) {
    perror("failed setting fcntls on socket");
    goto end;
  }

  status = fcntl(stream_lsocket, F_SETFL, flags | O_NONBLOCK);
  if (status == -1) {
    perror("failed setting fcntls on socket");
    goto end;
  }

  while (!dying) {
    /* get a stream connection, if one is waiting */
    starttime = time(NULL);
    b = sizeof(incoming);
    connected_socket = accept(stream_lsocket,
                              (struct sockaddr *)&incoming, &b);
    if (connected_socket >= 0) {
      conn = malloc( sizeof( connection ) );
      if (conn == NULL) {
        perror("");
        goto end;
      }

      init_connection(conn);
      conn->sock = connected_socket;
      conn->state = CONN_CONNECTED;
    }

    FD_ZERO(&r_set);
    FD_ZERO(&w_set);
    FD_ZERO(&e_set);
    max_fd = 0;

    cur_node = connections_root;
    while (cur_node != NULL) {
      assert(cur_node->this != NULL);
      conn = (connection *)cur_node->this;
      if (conn->state != CONN_CLOSED) {
        set_fdsets(conn, &max_fd, &r_set, &w_set, &e_set);
      }
      cur_node = cur_node->next;
    }

    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    num_fd = select(max_fd + 1, &r_set, &w_set, &e_set, &timeout);

    //printf("returned from select, max_fd: %d, num_fd: %d\n", max_fd, num_fd);

    cur_node = connections_root;
    while (cur_node != NULL) {
      assert(cur_node->this != NULL);
      conn = (connection *)cur_node->this;
      if (conn->state != CONN_CLOSED) {
        handle_socket_sets(conn, &r_set, &w_set, &e_set);
      }
      cur_node = cur_node->next;
    }

    cur_node = connections_root;
    while (cur_node != NULL) {
      assert(cur_node->this != NULL);
      conn = (connection *)cur_node->this;
      if (conn->state != CONN_CLOSED) {
        handle_connection(conn);
      }
      cur_node = cur_node->next;
    }

    free_closed_connections();
  }

 end:
  if (!dying) {
    fprintf(stderr, "critical failure in server thread\n");
    dying = TRUE;
  }
  return NULL;
}
